#include <asm/io.h>
#include <asm/system.h>
#include <display.h>
#include <kernel.h>
#include <irq.h>
#include <timer.h>
#include <syscall.h>
#include <task.h>

#define L_SHIFT            1
#define L_CTRL             2
#define L_ALT              4
#define R_ALT              8
#define R_CTRL            16
#define R_SHIFT           32

extern void irq_keyboard(void);
extern volatile int f1,f2;

static void kb_wait(void);
static unsigned int kb_wait_input(void);
static void send_cmd(unsigned char c);

static const unsigned char key_map[128] =
{
	0,0,'1','2','3','4','5','6',        /* 0x00 - 0x07 */
	'7','8','9','0','-','=',0,0,        /* 0x08 - 0x0f */
	'q','w','e','r','t','y','u','i',    /* 0x10 - 0x17 */
	'o','p','[',']','\n',0,'a','s',     /* 0x18 - 0x1f */
	'd','f','g','h','j','k','l',';',    /* 0x20 - 0x27 */
	'\'','`',0,'\\','z','x','c','v',    /* 0x28 - 0x2f */
	'b','n','m',',','.','/',0,0,        /* 0x30 - 0x37 */
	0,' ',0,0,0,0,0,0,                  /* 0x38 - 0x3f */
	0,0,0,0,0,0,0,0,                    /* 0x40 - 0x47 */
	0,0,0,0,0,0,0,0,                    /* 0x48 - 0x4f */
	0,0,0,0,0,0,0,0,                    /* 0x50 - 0x57 */
	0,0,0,0,0,0,0,0,                    /* 0x58 - 0x5f */
	0,0,0,0,0,0,0,0,                    /* 0x60 - 0x67 */
	0,0,0,0,0,0,0,0,                    /* 0x68 - 0x6f */
	0,0,0,0,0,0,0,0,                    /* 0x70 - 0x77 */
	0,0,0,0,0,0,0,0                     /* 0x78 - 0x7f */
};

//static unsigned int mode;

/*    */
void kb_wait(void)
{
	int i;

	for( i=0; i < 0x100000; i++ )
		if( (inb_p(0x64) & 0x02) == 0 )
			return;

	printk("\nKeyboard timed out\n");
}

unsigned int kb_wait_input(void)
{
	int i;

	i = 0x500000;
	do
	{
		if( (inb_p(0x64) & 0x01) == 0 )
			continue;
		return inb(0x60);
	} while( --i );

	return -1;
}


void send_cmd(unsigned char c)
{
	kb_wait();
	outb(c,0x64);
}

void do_keyboard(void)
{
	static unsigned char e = 0;
	unsigned char scancode;

	/*   */
	send_cmd(0xad);
	/*    */
	outb(0x20,0x20);

	scancode = inb(0x60);

	if( scancode == 0xe0 || scancode == 0xe1 )
	{
		e = scancode;
		goto end_kbd;
	}

	if( e )
	{
		printk(" 0x%x 0x%x ",e,scancode);
		e = 0;
		goto end_kbd;
	}

	if( scancode & 0x80 )
		goto end_kbd;

	if( key_map[scancode] )
	{
		putchar(key_map[scancode]);
		goto end_kbd;
	}
	else printk(" 0x%x ",scancode);

	if( scancode == 0x01 )
	{
		*((unsigned short *)0x472) = 0x1234;
		outb(0xfe,0x64);
		for(;;);
	}
	else if( scancode == 0x3b )
		timer_print();
	else if( scancode == 0x3c || scancode == 0x3d )
	{
		if( syscall(scancode-0x3b,0) == 0 )
			puts("Syscall return OK\n");
		else
			puts("Syscall return ERROR\n");
	}
	else if( scancode == 0x3e )
	{
		puts("\n*((char *)0xc0000) = 1;\n");
		*((char *)0xc0000) = 1;
	}
	else if( scancode == 0x3f )
		task_print();
	else if( scancode == 0x40 )
	{
		f1 = 1; f2 = 1;
	}

end_kbd:
	/*   */
	send_cmd(0xae);
}

void keyboard_init(void)
{
	kb_wait();
	outb_p(0xff,0x60);
	if( kb_wait_input() != 0xfa )
	{
		puts("Init kbd: reset kbd failed, not ACK.\n");
		return;
	}
	if( kb_wait_input() != 0xaa )
	{
		puts("Init kbd: reset kbd failed, not POR.\n");
		return;
	}

	set_irq(1,irq_keyboard);
	puts("Keyboard init OK...\n");
}
