#include <asm/io.h>
#include <kernel.h>
#include <display.h>

/*      */
struct regs
{
	unsigned long gs;
	unsigned long fs;
	unsigned long es;
	unsigned long ds;
	unsigned long edi;
	unsigned long esi;
	unsigned long esp;
	unsigned long ebp;
	unsigned long edx;
	unsigned long ecx;
	unsigned long ebx;
	unsigned long eax;
};

struct frame
{
	unsigned int error;
	unsigned int eip;
	unsigned int cs;
	unsigned int flags;
};

extern void divide_error(void);
extern void debug(void);
extern void nmi(void);
extern void int3(void);
extern void overflow(void);
extern void bound(void);
extern void invalid_op(void);
extern void not_device(void);
extern void double_fault(void);
extern void coprocessor_segment_ove(void);
extern void invalid_TSS(void);
extern void segment_not_present(void);
extern void segment_stack_error(void);
extern void general_protection(void);
extern void page_fault(void);
extern void coprocessor_error(void);
extern void trap_reserved(void);

struct desc idt[49] __attribute__ ((aligned (8)));

void set_trap(int i, void (* func)(void))
{
	/*     15 .. 0 */
	idt[i].a = (8 << 16) | ((unsigned int)func & 0xffff);
	/*      31 .. 16 */
	idt[i].b = ((unsigned int)func & 0xffff0000) | 0x8f00;
}

static void print_regs(const struct regs *reg, const struct frame *frm)
{
	printk("EAX = %p; EBX = %p; ECX = %p; EDX = %p;\n"
			"EBP = %p; ESP = %p; ESI = %p; EDI = %p;\n"
			"DS = %p; ES = %p; FS = %p; GS = %p;\n"
			"eip = %p; cs = %p; flags = %p; error = %x\n",
			reg->eax,reg->ebx,reg->ecx,reg->edx,
			reg->ebp,reg->esp,reg->esi,reg->edi,
			reg->ds,reg->es,reg->fs,reg->gs,
			frm->eip,frm->cs,frm->flags,frm->error);
}

void do_divide_error(const struct regs *reg, const struct frame *frm)
{
	puts("trap 0: divide error\n");
	print_regs(reg,frm);
	for(;;);
}

void do_debug(const struct regs *reg, const struct frame *frm)
{
	puts("trap 1: debug\n");
	print_regs(reg,frm);
	for(;;);
}

void do_nmi(const struct regs *reg, const struct frame *frm)
{
	puts("trap 2: nmi\n");
	print_regs(reg,frm);
	for(;;);
}

void do_int3(const struct regs *reg, const struct frame *frm)
{
	puts("trap 3: int3\n");
	print_regs(reg,frm);
	for(;;);
}

void do_overflow(const struct regs *reg, const struct frame *frm)
{
	puts("trap 4: overflow\n");
	print_regs(reg,frm);
	for(;;);
}

void do_bound(const struct regs *reg, const struct frame *frm)
{
	puts("trap 5: bound\n");
	print_regs(reg,frm);
	for(;;);
}

void do_invalid_op(const struct regs *reg, const struct frame *frm)
{
	puts("trap 6: invalid operand\n");
	print_regs(reg,frm);
	for(;;);
}

void do_not_device(const struct regs *reg, const struct frame *frm)
{
	puts("trap 7: not device\n");
	print_regs(reg,frm);
	for(;;);
}

void do_double_fault(const struct regs *reg, const struct frame *frm)
{
	puts("trap 8: double fault\n");
	print_regs(reg,frm);
	for(;;);
}

void do_coprocessor_segment_ove(const struct regs *reg, const struct frame *frm)
{
	puts("trap 9: coprocessor segmen overrun\n");
	print_regs(reg,frm);
	for(;;);
}

void do_invalid_TSS(const struct regs *reg, const struct frame *frm)
{
	puts("trap 10: invalid TSS\n");
	print_regs(reg,frm);
	for(;;);
}

void do_segment_not_present(const struct regs *reg, const struct frame *frm)
{
	puts("trap 11: segmen not present\n");
	print_regs(reg,frm);
	for(;;);
}

void do_segment_stack_error(const struct regs *reg, const struct frame *frm)
{
	puts("trap 12: segmen stack error\n");
	print_regs(reg,frm);
	for(;;);
}

void do_general_protection(const struct regs *reg, const struct frame *frm)
{
	puts("trap 13: general protection\n");
	print_regs(reg,frm);
	for(;;);
}

void do_page_fault(const struct regs *reg, const struct frame *frm)
{
	puts("trap 14: page fault\n");
	print_regs(reg,frm);
	for(;;);
}

void do_coprocessor_error(const struct regs *reg, const struct frame *frm)
{
	puts("trap 16: coprocessor error\n");
	print_regs(reg,frm);
	for(;;);
}

void do_trap_reserved(const struct regs *reg, const struct frame *frm)
{
	puts("trap reserved\n");
	print_regs(reg,frm);
	for(;;);
}

void traps_init(void)
{
	int i;

	set_trap(0,divide_error);
	set_trap(1,debug);
	set_trap(2,nmi);
	set_trap(3,int3);
	set_trap(4,overflow);
	set_trap(5,bound);
	set_trap(6,invalid_op);
	set_trap(7,not_device);
	set_trap(8,double_fault);
	set_trap(9,coprocessor_segment_ove);
	set_trap(10,invalid_TSS);
	set_trap(11,segment_not_present);
	set_trap(12,segment_stack_error);
	set_trap(13,general_protection);
	set_trap(14,page_fault);
	set_trap(15,trap_reserved);
	set_trap(16,coprocessor_error);
	for( i=17; i < 32; i++ )
		set_trap(i,trap_reserved);

	puts("Traps init OK...\n");
}
